# stdlib
import abc
import os


class _Client(abc.ABC):
    def __init__(
        self,
        *,
        api_key: str = None,
        service_token: str = None,
        host: str = None,
    ):
        self.api_key = api_key or os.getenv('DBT_CLOUD_API_KEY', None)
        self.service_token = service_token or os.getenv('DBT_CLOUD_SERVICE_TOKEN', None)
        self._host = host or os.getenv('DBT_CLOUD_HOST', self._default_domain)

    @property
    @abc.abstractmethod
    def _default_domain(self):
        """Default used when host is not provided by a user"""

    @property
    @abc.abstractmethod
    def _path(self):
        """Path where specific resource is located"""

    @property
    @abc.abstractmethod
    def _header_property(self):
        """Define what property to use within the headers passed with a request"""

    @property
    def _base_url(self):
        return f'https://{self._host}{self._path}'

    @property
    def headers(self):
        return {
            'Authorization': f'Token {getattr(self, self._header_property)}',
            'Content-Type': 'application/json',
        }

    def full_url(self, path: str = None):
        if path is not None:
            return f'{self._base_url}{path}'

        return self._base_url



# first party
from dbtc.client.cloud import _CloudClient
from dbtc.client.metadata import _MetadataClient


class dbtCloudClient:
    def __init__(self, **kwargs):
        self.cloud = _CloudClient(**kwargs)
        self.metadata = _MetadataClient(**kwargs)



# stdlib
import enum
import time
from functools import partial, wraps
from typing import Dict, List

# third party
import requests

# first party
from dbtc.client.base import _Client


def _version_decorator(func, version):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self._path = f'/api/{version}/'
        result = func(self, *args, **kwargs)
        return result

    return wrapper


v2 = partial(_version_decorator, version='v2')
v3 = partial(_version_decorator, version='v3')
v4 = partial(_version_decorator, version='v4')


class JobRunStatus(enum.IntEnum):
    QUEUED = 1
    STARTING = 2
    RUNNING = 3
    SUCCESS = 10
    ERROR = 20
    CANCELLED = 30


class _CloudClient(_Client):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.session = requests.Session()
        self.session.headers = self.headers

    _default_domain = 'cloud.getdbt.com'
    _path = None

    @property
    def _header_property(self):
        if self.api_key is None:
            return 'service_token'

        return 'api_key'

    def _make_request(
        self, path: str, *, method: str = 'get', **kwargs
    ) -> requests.Response:
        """Make request to API."""
        full_url = self.full_url(path)
        response = self.session.request(method=method, url=full_url, **kwargs)
        return response

    def _simple_request(self, path: str, *, method: str = 'get', **kwargs) -> Dict:
        """Return json from response."""
        response = self._make_request(path, method=method, **kwargs)
        return response.json()

    def _paginated_request(
        self, path: str, *, method: str = 'get', **kwargs
    ) -> List[Dict]:
        """Multiple paginated requests given presence of specific header.

        Note:  Only available in V4.
        """
        response = self._make_request(path, method=method, **kwargs)
        data = []
        while True:
            response_data = response.json().get('data', [])
            data.extend(response_data)
            next_page_token = self._get_pagination_token(response)
            if next_page_token is not None:
                response = self._make_request(
                    path,
                    method=method,
                    headers={'x-dbt-continuation-token': next_page_token},
                    **kwargs,
                )
            else:
                break
        return data

    def _get_pagination_token(self, response):
        """Retrieve pagination token.

        Note:  Only available in V4.
        """
        return response.headers.get('x-dbt-continuation-token', None)

    def _get_by_name(self, items: List, item_name: str, value: str = 'name'):
        try:
            obj = [item for item in items if item[value] == item_name][0]
        except IndexError:
            obj = None
        return obj

    @v3
    def assign_group_permissions(
        self, account_id: int, group_id: int, payload: Dict
    ) -> Dict:
        """Assign group permissions

        Args:
            account_id (int): Numeric ID of the account
            group_id (int): Numeric ID of the group
            payload (dict): Dictionary representing the group to create
        """
        return self._simple_request(
            f'accounts/{account_id}/group-permissions/{group_id}/',
            method='post',
            json=payload,
        )

    @v3
    def assign_service_token_permissions(
        self, account_id: int, service_token_id: int, payload: List[Dict]
    ) -> Dict:
        """Assign permissions to a service token.

        Args:
            account_id (int): Numeric ID of the account
            service_token_id (int): Numeric ID of the service token
            payload (list): List of dictionaries representing the permissions to assign
        """
        return self._simple_request(
            f'accounts/{account_id}/service-tokens/{service_token_id}/permissions/',
            method='post',
            json=payload,
        )

    @v3
    def assign_user_to_group(
        self, account_id: int, project_id: int, payload: Dict
    ) -> Dict:
        """Assign a user to a group

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            payload (dict): Dictionary representing the user to assign
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/assign-groups/',
            method='post',
            json=payload,
        )

    @v2
    def cancel_run(self, account_id: int, run_id: int):
        """Cancel a run.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            run_id (int): Numeric ID of the run to retrieve
        """
        return self._simple_request(
            f'accounts/{account_id}/runs/{run_id}/cancel',
            method='post',
        )

    @v3
    def create_adapter(self, account_id: int, project_id: int, payload: Dict) -> Dict:
        """Create an adapter

        Note:  This is a prerequisite for creating a Databricks connection

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            payload (dict): Dictionary representing the adapter to create
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/adapters/',
            method='post',
            json=payload,
        )

    @v3
    def create_connection(
        self, account_id: int, project_id: int, payload: Dict
    ) -> Dict:
        """Create a connection

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            payload (dict): Dictionary representing the connection to create
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/connections/',
            method='post',
            json=payload,
        )

    @v3
    def create_credentials(
        self, account_id: int, project_id: int, payload: Dict
    ) -> Dict:
        """Create credentials

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            payload (dict): Dictionary representing the credentials to create
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/credentials/',
            method='post',
            json=payload,
        )

    @v3
    def create_environment(
        self, account_id: int, project_id: int, payload: Dict
    ) -> Dict:
        """Create an environment

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            payload (dict): Dictionary representing the environment to create
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/environments/',
            method='post',
            json=payload,
        )

    @v3
    def create_environment_variables(
        self, account_id: int, project_id: int, payload: Dict
    ) -> Dict:
        """Create environment variabless

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            payload (dict): Dictionary representing the environment variables to create
        """
        url = f'accounts/{account_id}/projects/{project_id}/environment-variables/'
        if len(payload.keys()) > 1:
            url += 'bulk/'
        return self._simple_request(url, method='post', json=payload)

    @v3
    def create_job(self, account_id: int, project_id: int, payload: Dict) -> Dict:
        """Create a job

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            payload (dict): Dictionary representing the job to create
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/jobs/',
            method='post',
            json=payload,
        )

    @v3
    def create_project(self, account_id: int, payload: Dict) -> Dict:
        """Create a project

        Args:
            account_id (int): Numeric ID of the account
            payload (dict): Dictionary representing the project to create
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/', method='post', json=payload
        )

    @v3
    def create_repository(
        self, account_id: int, project_id: int, payload: Dict
    ) -> Dict:
        """Create a repository

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            payload (dict): Dictionary representing the repository to create
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/repositories/',
            method='post',
            json=payload,
        )

    @v3
    def create_service_token(self, account_id: int, payload: Dict) -> Dict:
        """Create a service token

        Args:
            account_id (int): Numeric ID of the account
            payload (dict): Dictionary representing the service token to create
        """
        return self._simple_request(
            f'accounts/{account_id}/service-tokens/', method='post', json=payload
        )

    @v3
    def create_user_group(self, account_id: int, payload: Dict) -> Dict:
        """Create a user group

        Args:
            account_id (int): Numeric ID of the account
            payload (dict): Dictionary representing the group to create
        """
        return self._simple_request(
            f'accounts/{account_id}/groups/', method='post', json=payload
        )

    @v3
    def deactivate_user_license(
        self, account_id: int, permission_id: int, payload: Dict
    ) -> Dict:
        """Deactivate user license

        Args:
            account_id (int): Numeric ID of the account
            permission_id (int): Numeric ID of the permission that contains
                user you'd like to deactivate
        """
        return self._simple_request(
            f'accounts/{account_id}/permissions/{permission_id}',
            method='post',
            json=payload,
        )

    @v3
    def delete_connection(
        self, account_id: int, project_id: int, connection_id: int
    ) -> Dict:
        """Delete a connection

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            connection_id (int): Numeric ID of the connection to delete
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/connections/{connection_id}',
            method='delete',
        )

    @v3
    def delete_environment(self, account_id: int, environment_id: int) -> Dict:
        """Delete job for a specified account

        Args:
            account_id (int): Numeric ID of the account
            environment_id (int): Numeric ID of the environment to delete
        """
        return self._simple_request(
            f'accounts/{account_id}/environments/{environment_id}/',
            method='delete',
        )

    @v3
    def delete_environment_variables(
        self, account_id: int, project_id: int, payload: Dict
    ) -> Dict:
        """Delete environment variables for a specified account

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            payload (Dict): Dictionary representing environment variables to delete
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/environment-variables/bulk/',
            method='delete',
            json=payload,
        )

    @v3
    def delete_job(self, account_id: int, job_id: int) -> Dict:
        """Delete job for a specified account

        Args:
            account_id (int): Numeric ID of the account
            job_id (int): Numeric ID of the project to delete
        """
        return self._simple_request(
            f'accounts/{account_id}/jobs/{job_id}/',
            method='delete',
        )

    @v3
    def delete_project(self, account_id: int, project_id: int) -> Dict:
        """Delete project for a specified account

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project to delete
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/',
            method='delete',
        )

    @v3
    def delete_repository(
        self, account_id: int, project_id: int, repository_id: int
    ) -> Dict:
        """Delete repository for a specified account

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            repository_id (int): Numeric ID of the repository to delete
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/repositories/{repository_id}',
            method='delete',
        )

    @v3
    def delete_user_group(self, account_id: int, group_id: int) -> Dict:
        """Delete group for a specified account

        Args:
            account_id (int): Numeric ID of the account
            group_id (int): Numeric ID of the group to delete
        """
        return self._simple_request(
            f'accounts/{account_id}/groups/{group_id}/',
            method='post',
        )

    @v2
    def get_account(self, account_id: int) -> Dict:
        """Get an account by its ID.

        Args:
            account_id (int): Numeric ID of the account to retrieve
        """
        return self._simple_request(f'accounts/{account_id}')

    @v2
    def get_account_by_name(self, account_name: str) -> Dict:
        """Get an account by its name.

        Args:
            account_name (str): Name of an account
        """
        accounts = self.list_accounts()
        account = self._get_by_name(accounts['data'], account_name)
        if account is not None:
            return self.get_account(account['id'])

        raise Exception(f'"{account_name}" was not found')

    @v2
    def get_account_licenses(self, account_id: int) -> Dict:
        """List account licenses for a specified account.

        Args:
            account_id (int): Numeric ID of the account to retrieve
        """
        return self._simple_request(f'accounts/{account_id}/licenses')

    @v2
    def get_job(self, account_id: int, job_id: int, *, order_by: str = None) -> Dict:
        """Get a job by its ID.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            job_id (int): Numeric ID of the job to retrieve
            order_by (:obj:`str`, optional): Field to order the result by.
                Use - to indicate reverse order.
        """
        return self._simple_request(
            f'accounts/{account_id}/jobs/{job_id}/',
            params={'order_by': order_by},
        )

    @v2
    def get_project(self, account_id: int, project_id: int) -> Dict:
        """Get a project by its ID.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            project_id (int): Numeric ID of the project to retrieve
        """
        return self._simple_request(f'accounts/{account_id}/projects/{project_id}')

    @v2
    def get_project_by_name(
        self, project_name: str, account_id: int = None, account_name: str = None
    ) -> Dict:
        """Get a project by its name.

        Args:
            project_name (str): Name of project to retrieve
            account_id (:obj:`int`, optional): Numeric ID of the account to retrieve
            account_name (:obj:`str`, optional): Name of account to retrieve
        """
        if account_id is None and account_name is None:
            accounts = self.list_accounts()
            for account in accounts['data']:
                projects = self.list_projects(account['id'])
                project = self._get_by_name(projects['data'], project_name)
                if project is not None:
                    break

        else:
            if account_id is not None:
                account = self.get_account(account_id)
            else:
                account = self.get_account_by_name(account_name)
            projects = self.list_projects(account['id'])
            project = self._get_by_name(projects['data'], project_name)

        if project is not None:
            return self.get_project(project['account_id'], project['id'])

        raise Exception(f'"{project_name}" was not found.')

    @v2
    def get_run(
        self, account_id: int, run_id: int, *, include_related: List[str] = None
    ) -> Dict:
        """Get a run by its ID.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            run_id (int): Numeric ID of the run to retrieve
            include_related (:obj:`list` of :obj:`str`, optional): List of related
                fields to pull with the run. Valid values are "trigger", "job",
                "repository", "debug_logs", "run_steps", and "environment".
        """
        return self._simple_request(
            f'accounts/{account_id}/runs/{run_id}',
            params={'include_related': ','.join(include_related or [])},
        )

    @v2
    def get_run_artifact(
        self,
        account_id: int,
        run_id: int,
        path: str,
        *,
        step: int = None,
    ) -> Dict:
        """Fetch artifacts from a completed run.

        Once a run has completed, you can use this endpoint to download the
        manifest.json, run_results.json or catalog.json files from dbt Cloud. These
        artifacts contain information about the models in your dbt project, timing
        information around their execution, and a status message indicating the result
        of the model build.

        Note: By default, this endpoint returns artifacts from the last step in the
        run. To list artifacts from other steps in the run, use the step query
        parameter described below.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            run_id (int): Numeric ID of the run to retrieve
            path (str): Paths are rooted at the target/ directory. Use manifest.json,
                catalog.json, or run_results.json to download dbt-generated artifacts
                for the run.
            step (:obj:`str`, optional): The index of the Step in the Run to query for
                artifacts. The first step in the run has the index 1. If the step
                parameter is omitted, then this endpoint will return the artifacts
                compiled for the last step in the run.
        """
        return self._simple_request(
            f'accounts/{account_id}/runs/{run_id}/artifacts/{path}',
            params={'step': step},
        )

    @v3
    def get_run_timing_details(
        self, account_id: int, project_id: int, run_id: int
    ) -> Dict:
        """Retrieves the timing details related to a run

        Args:
            account_id (int): Numeric ID of the account to retrieve
            run_id (int): Numeric ID of the run to retrieve
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/runs/{run_id}/timing/'
        )

    @v4
    def get_run_v4(self, account_id: int, run_id: int) -> Dict:
        """Retrieves the details of an existing run with the given run_id.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            run_id (int): Numeric ID of the run to retrieve
        """
        return self._simple_request(f'accounts/{account_id}/runs/{run_id}')

    @v3
    def get_service_token(self, account_id: int, service_token_id: int) -> Dict:
        """Retrieves a service token.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            service_token_id (int): Numeric ID of the service token to retrieve
        """
        return self._simple_request(
            f'accounts/{account_id}/service-tokens/{service_token_id}'
        )

    @v2
    def get_user(self, account_id: int, user_id: int) -> Dict:
        """List invited users in an account.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            user_id (int): Numeric ID of the user to retrieve
        """
        return self._simple_request(f'accounts/{account_id}/users/{user_id}/')

    @v2
    def list_accounts(self) -> Dict:
        """List of accounts that your API Token is authorized to access."""
        return self._simple_request('accounts/')

    @v3
    def list_audit_logs(
        self,
        account_id: int,
        *,
        logged_at_start: str = None,
        logged_at_end: str = None,
        offset: int = None,
        limit: int = None,
    ) -> Dict:
        """List audit logs for a specific account

        Args:
            account_id (int): Numeric ID of the account to retrieve
            logged_at_start (:obj:`str`, optional):  Date to begin retrieving audit
                logs
                Format is yyyy-mm-dd
            logged_at_end (:obj:`str`, optional): Date to stop retrieving audit logs.
                Format is yyyy-mm-dd
            offset (:obj:`int`, optional): The offset to apply when listing runs.
                Use with limit to paginate results.
            limit (:obj:`int`, optional): The limit to apply when listing runs.
                Use with offset to paginate results.
        """
        return self._simple_request(
            f'accounts/{account_id}/audit-logs',
            params={
                'logged_at_start': logged_at_start,
                'logged_at_end': logged_at_end,
                'offset': offset,
                'limit': limit,
            },
        )

    @v3
    def list_connections(self, account_id: int, project_id: int) -> Dict:
        """List connections for a specific account and project

        Args:
            account_id (int): Numeric ID of the account to retrieve
            project_id (int): Numeric ID of the project to retrieve
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/connections'
        )

    @v3
    def list_credentials(self, account_id: int, project_id: int) -> Dict:
        """List credentials for a specific account and project

        Args:
            account_id (int): Numeric ID of the account to retrieve
            project_id (int): Numeric ID of the project to retrieve
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/credentials'
        )

    @v3
    def list_environments(self, account_id: int, project_id: int) -> Dict:
        """List environments for a specific account and project

        Args:
            account_id (int): Numeric ID of the account to retrieve
            project_id (int): Numeric ID of the project to retrieve
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/environments/'
        )

    @v3
    def list_feature_flags(self, account_id: int) -> Dict:
        """List feature flags for a specific account

        Args:
            account_id (int): Numeric ID of the account to retrieve
        """
        return self._simple_request(f'accounts/{account_id}/feature-flag/')

    @v3
    def list_groups(self, account_id: int) -> Dict:
        """List groups for a specific account and project

        Args:
            account_id (int): Numeric ID of the account to retrieve
            project_id (int): Numeric ID of the project to retrieve
        """
        return self._simple_request(f'accounts/{account_id}/groups/')

    @v2
    def list_invited_users(self, account_id: int) -> Dict:
        """List invited users in an account.

        Args:
            account_id (int): Numeric ID of the account to retrieve
        """
        return self._simple_request(f'accounts/{account_id}/invites/')

    @v2
    def list_jobs(
        self, account_id: int, *, order_by: str = None, project_id: int = None
    ) -> Dict:
        """List jobs in an account or specific project.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            order_by (:obj:`str`, optional): Field to order the result by.
                Use - to indicate reverse order.
            project_id (:obj:`int`, optional): Numeric ID of the project containing jobs
        """
        return self._simple_request(
            f'accounts/{account_id}/jobs/',
            params={'order_by': order_by, 'project_id': project_id},
        )

    @v3
    def list_projects(self, account_id: int) -> Dict:
        """List projects for a specified account.

        Args:
            account_id (int): Numeric ID of the account to retrieve
        """
        return self._simple_request(f'accounts/{account_id}/projects')

    @v3
    def list_repositories(self, account_id: int, project_id: int) -> Dict:
        """List repositories for a specific account and project

        Args:
            account_id (int): Numeric ID of the account to retrieve
            project_id (int): Numeric ID of the project to retrieve
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/repositories/'
        )

    @v2
    def list_run_artifacts(
        self,
        account_id: int,
        run_id: int,
        *,
        step: int = None,
    ) -> Dict:
        """Fetch a list of artifact files generated for a completed run.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            run_id (int): Numeric ID of the run to retrieve
            step (:obj:`str`, optional): The index of the Step in the Run to query for
                artifacts. The first step in the run has the index 1. If the step
                parameter is omitted, then this endpoint will return the artifacts
                compiled for the last step in the run.
        """
        return self._simple_request(
            f'accounts/{account_id}/runs/{run_id}/artifacts',
            params={'step': step},
        )

    @v2
    def list_runs(
        self,
        account_id: int,
        *,
        include_related: List[str] = None,
        job_definition_id: int = None,
        order_by: str = None,
        offset: int = None,
        limit: int = None,
    ) -> Dict:
        """List runs in an account.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            include_related (:obj:`list` of :obj:`str`, optional): List of related
                fields to pull with the run. Valid values are "trigger", "job",
                "repository", "debug_logs", "run_steps", and "environment".
            job_definition_id (:obj:`int`, optional): Applies a filter to only return
                runs
                from the specified Job.
            order_by (:obj:`str`, optional): Field to order the result by.
                Use - to indicate reverse order.
            offset (:obj:`int`, optional): The offset to apply when listing runs.
                Use with limit to paginate results.
            limit (:obj:`int`, optional): The limit to apply when listing runs.
                Use with offset to paginate results.
        """
        return self._simple_request(
            f'accounts/{account_id}/runs',
            params={
                'include_related': ','.join(include_related or []),
                'job_definition_id': job_definition_id,
                'order_by': order_by,
                'offset': offset,
                'limit': limit,
            },
        )

    @v4
    def list_runs_v4(
        self,
        account_id: int,
        *,
        limit: int = None,
        environment: str = None,
        project: str = None,
        job: str = None,
        status: str = None,
    ) -> List[Dict]:
        """Returns a list of runs in the account.

        The runs are returned sorted by creation date, with the most recent run
        appearing first.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            limit (:obj:`int`, optional): A limit on the number of objects to be
                returned, between 1 and 100.
            environment (str): A filter on the list based on the object's
                environment_id field.
            project (str): A filter on the list based on the object's project_id field.
            job (str): A filter on the list based on the object's job_id field.
            status: A filter on the list based on the object's status field.
                Enum: "Queued" "Starting" "Running" "Succeeded" "Failed" "Canceled"
        """
        return self._paginated_request(
            f'accounts/{account_id}/runs',
            params={
                'limit': limit,
                'environment': environment,
                'project': project,
                'job': job,
                'status': status,
            },
        )

    @v3
    def list_service_token_permissions(
        self, account_id: int, service_token_id: int
    ) -> Dict:
        """List service token permissions for a specific account.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            service_token_id (int): Numeric ID of the service token to retrieve
        """
        return self._simple_request(
            f'accounts/{account_id}/service-tokens/{service_token_id}/permissions'
        )

    @v3
    def list_service_tokens(self, account_id: int) -> Dict:
        """List service tokens for a specific account.

        Args:
            account_id (int): Numeric ID of the account to retrieve
        """
        return self._simple_request(f'accounts/{account_id}/service-tokens/')

    @v2
    def list_users(
        self,
        account_id: int,
        *,
        limit: int = None,
        offset: int = None,
        order_by: str = 'email',
    ) -> Dict:
        """List users in an account.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            limit (:obj:`int`, optional): The limit to apply when listing runs.
                Use with offset to paginate results.
            offset (:obj:`int`, optional): The offset to apply when listing runs.
                Use with limit to paginate results.
            order_by (:obj:`str`, optional): Field to order the result by.
                Use - to indicate reverse order.
        """
        return self._simple_request(
            f'accounts/{account_id}/users/',
            params={'limit': limit, 'offset': offset, 'order_by': order_by},
        )

    @v3
    def test_connection(self, account_id: int, payload: Dict) -> Dict:
        """Test a connection

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            payload (dict): Dictionary representing the connection to test
        """
        return self._simple_request(
            f'accounts/{account_id}/connections/test/', method='post', json=payload
        )

    @v2
    def trigger_job(self, account_id: int, job_id: int, payload: Dict):
        """Trigger a job by its ID.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            job_id (int): Numeric ID of the job to trigger
            payload (dict): Payload required for post request
        """
        return self._simple_request(
            f'accounts/{account_id}/jobs/{job_id}/run/',
            method='post',
            json=payload,
        )

    @v2
    def trigger_job_and_poll(
        self, account_id: int, job_id: int, payload: Dict, poll_interval: int = 10
    ) -> int:
        """Trigger a job by its ID and poll until completion:  one of
          SUCCESS, ERROR, or CANCELLED.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            job_id (int): Numeric ID of the job to trigger
            payload (dict): Payload required for post request
            poll_interval (:obj:`int`, optional): Number of seconds to wait in between
                polling
        """

        def _run_status_formatted(run_id: int, status: str, time: float) -> str:
            """Format a string indicating status of job.

            Args:
                run_id (int): Numeric ID of the run to retrieve
                status (str): Status of job
                time (float): Elapsed time since job triggered
            """
            return (
                f'Run {run_id} - {status.capitalize()}, Elapsed time: {round(time, 0)}s'
            )

        run_id = self.trigger_job(account_id, job_id, payload)['data']['id']
        print('Job Triggered!')
        start = time.time()

        while True:
            time.sleep(poll_interval)
            run = self.get_run(account_id, run_id)
            status = run['data']['status']
            status_name = JobRunStatus(status).name
            if status == JobRunStatus.SUCCESS:
                print(_run_status_formatted(run_id, status_name, time.time() - start))
                return run_id
            if status in [JobRunStatus.CANCELLED, JobRunStatus.ERROR]:
                raise Exception(run['data']['status_message'])
            print(_run_status_formatted(run_id, status_name, time.time() - start))

    @v3
    def update_connection(
        self, account_id: int, project_id: int, connection_id: int, payload: Dict
    ) -> Dict:
        """Update a connection

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            connection_id (int): Numeric ID of the connection to update
            payload (dict): Dictionary representing the connection to update
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/connections/{connection_id}/',
            method='post',
            json=payload,
        )

    @v3
    def update_credentials(
        self, account_id: int, project_id: int, credentials_id: int, payload: Dict
    ) -> Dict:
        """Update credentials

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            credentials_id (int): Numeric ID of the credentials to update
            payload (dict): Dictionary representing the credentials to update
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/credentials/{credentials_id}/',  # noqa: E50
            method='post',
            json=payload,
        )

    @v3
    def update_environment(
        self, account_id: int, project_id: int, environment_id: int, payload: Dict
    ) -> Dict:
        """Update a connection

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            environment_id (int): Numeric ID of the environment to update
            payload (dict): Dictionary representing the environment to update
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/environments/{environment_id}/',  # noqa: E501
            method='post',
            json=payload,
        )

    @v2
    def update_job(self, account_id: int, job_id: int, payload: Dict) -> Dict:
        """Update a job by its ID.

        Args:
            account_id (int): Numeric ID of the account to retrieve
            job_id (int): Numeric ID of the job to retrieve
            payload (dict): Payload required for post request
        """
        return self._simple_request(
            f'accounts/{account_id}/jobs/{job_id}/',
            method='post',
            json=payload,
        )

    @v3
    def update_project(self, account_id: int, project_id: int, payload: Dict) -> Dict:
        """Update project for a specified account

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project to update
            payload (dict): Dictionary representing the project to update
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/', method='POST', json=payload
        )

    @v3
    def update_repository(
        self, account_id: int, project_id: int, repository_id: int, payload: Dict
    ) -> Dict:
        """Update a connection

        Args:
            account_id (int): Numeric ID of the account
            project_id (int): Numeric ID of the project
            repository_id (int): Numeric ID of the repository to update
            payload (dict): Dictionary representing the repository to update
        """
        return self._simple_request(
            f'accounts/{account_id}/projects/{project_id}/repositories/{repository_id}/',  # noqa: E501
            method='post',
            json=payload,
        )



# stdlib
import json
from typing import List, Optional

# third party
import typer

# first party
from dbtc import dbtCloudClient as dbtc

app = typer.Typer()


valid_inclusions = ['trigger', 'environment', 'run_steps', 'job', 'repository']


def complete_inclusion(ctx, param, incomplete):
    for inclusion in valid_inclusions:
        if inclusion.startswith(incomplete):
            yield inclusion


ACCOUNT_ID = typer.Option(
    ...,
    '--account-id',
    '-a',
    envvar='DBT_CLOUD_ACCOUNT_ID',
    help='Numeric ID of account to retrieve.',
)
API_KEY = typer.Option(
    None, '--api-key', envvar='DBT_CLOUD_API_KEY', help="User's dbt Cloud API Key"
)
CONNECTION_ID = typer.Option(
    ..., '--connection-id', '-c', help='Numeric ID of the connection.'
)
ENVIRONMENT_ID = typer.Option(
    ..., '--environment-id', '-e', help='Numeric ID of the connection.'
)
GROUP_ID = typer.Option(
    ..., '--group-id', '-g', help='Numeric ID of the group to retrieve.'
)
HOST = typer.Option(
    None,
    '--host',
    envvar='DBT_CLOUD_HOST',
    help='Only used for single tenant instances.',
)
INCLUDE_RELATED = typer.Option(
    None,
    '--include-related',
    '-i',
    help='List of related fields to pull with run',
    shell_complete=complete_inclusion,
)
JOB_ID = typer.Option(..., '--job-id', '-j', help='Numeric ID of job to retrieve.')
LIMIT = typer.Option(
    None,
    help='Limit to apply when listing runs.  Use with `offset` to paginate results.',
)
OFFSET = typer.Option(
    None,
    help='Offset to apply when listing runs.  Use with `limit` to paginate results.',
)
ORDER_BY = typer.Option(
    None,
    '--order-by',
    '-o',
    help='Field to order the result by.  Use `-` to indicate reverse order.',
)
PAYLOAD = typer.Option(
    ...,
    '--payload',
    '-d',
    help='String representation of dictionary needed to create or update resource.',
)
PROJECT_ID = typer.Option(
    ...,
    '--project-id',
    '-p',
    envvar='DBT_CLOUD_PROJECT_ID',
    help='Numeric ID of the project to retrieve.',
)
REPOSITORY_ID = typer.Option(
    ..., '--repository-id', '-y', help='Numeric ID of the repository.'
)
RUN_ID = typer.Option(..., '--run-id', '-r', help='Numeric ID of run to retrieve.')
SERVICE_TOKEN_ID = typer.Option(
    ..., '--service-token-id', '-t', help='Numeric ID of the service token.'
)
TOKEN = typer.Option(
    None,
    '--token',
    envvar='DBT_CLOUD_SERVICE_TOKEN',
    help='Service token for dbt Cloud Account.',
)
USER_ID = typer.Option(..., '--user-id', '-u', help='Numeric ID of the user.')


def _dbt_cloud_request(ctx: typer.Context, method: str, *args, **kwargs):
    data = getattr(dbtc(**ctx.obj).cloud, method)(*args, **kwargs)
    typer.echo(json.dumps(data))
    return data


@app.callback()
def common(
    ctx: typer.Context,
    api_key: Optional[str] = API_KEY,
    service_token: Optional[str] = TOKEN,
    host: Optional[str] = HOST,
):
    ctx.obj = ctx.params
    pass


@app.command()
def assign_group_permissions(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    group_id: int = GROUP_ID,
    payload: str = PAYLOAD,
):
    """Assign group permissions."""
    _dbt_cloud_request(
        ctx,
        'assign_group_permissions',
        account_id,
        group_id,
        json.loads(payload),
    )


@app.command()
def assign_service_token_permissions(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    service_token_id: int = SERVICE_TOKEN_ID,
    payload: str = PAYLOAD,
):
    """Assign permissions to a service token."""
    _dbt_cloud_request(
        ctx,
        'assign_service_token_permissions',
        account_id,
        service_token_id,
        json.loads(payload),
    )


@app.command()
def assign_user_to_group(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Assign a user to a group."""
    _dbt_cloud_request(
        ctx,
        'assign_user_to_group',
        account_id,
        project_id,
        json.loads(payload),
    )


@app.command()
def cancel_run(ctx: typer.Context, account_id: int = ACCOUNT_ID, run_id: int = RUN_ID):
    """Cancel a run."""
    _dbt_cloud_request(ctx, 'cancel_run', account_id, run_id)


@app.command()
def create_adapter(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Create an adapter."""
    _dbt_cloud_request(
        ctx, 'create_adapter', account_id, project_id, json.loads(payload)
    )


@app.command()
def create_connection(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Create a connection."""
    _dbt_cloud_request(
        ctx, 'create_connection', account_id, project_id, json.loads(payload)
    )


@app.command()
def create_credentials(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Create credentials."""
    _dbt_cloud_request(
        ctx, 'create_credentials', account_id, project_id, json.loads(payload)
    )


@app.command()
def create_environment(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Create an environment."""
    _dbt_cloud_request(
        ctx, 'create_environment', account_id, project_id, json.loads(payload)
    )


@app.command()
def create_environment_variables(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Create environment variables."""
    _dbt_cloud_request(
        ctx, 'create_environment_variables', account_id, project_id, json.loads(payload)
    )


@app.command()
def create_job(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Create a job in a project."""
    _dbt_cloud_request(ctx, 'create_job', account_id, project_id, json.loads(payload))


@app.command()
def create_project(
    ctx: typer.Context, account_id: int = ACCOUNT_ID, payload: str = PAYLOAD
):
    """Create a project."""
    _dbt_cloud_request(ctx, 'create_project', account_id, json.loads(payload))


@app.command()
def create_repository(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Create a repository in a project."""
    _dbt_cloud_request(
        ctx, 'create_repository', account_id, project_id, json.loads(payload)
    )


@app.command()
def create_service_token(
    ctx: typer.Context, account_id: int = ACCOUNT_ID, payload: str = PAYLOAD
):
    """Create a service token."""
    _dbt_cloud_request(ctx, 'create_service_token', account_id, json.loads(payload))


@app.command()
def create_user_group(
    ctx: typer.Context, account_id: int = ACCOUNT_ID, payload: str = PAYLOAD
):
    """Create a user group."""
    _dbt_cloud_request(ctx, 'create_user_group', account_id, json.loads(payload))


@app.command()
def deactivate_user_license(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    permission_id: int = typer.Option(..., '--permission-id'),
    payload: str = PAYLOAD,
):
    """Deactive a user license."""
    _dbt_cloud_request(
        ctx, 'deactivate_user_license', account_id, permission_id, json.loads(payload)
    )


@app.command()
def delete_connection(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    connection_id: int = CONNECTION_ID,
):
    """Delete a connection"""
    _dbt_cloud_request(ctx, 'delete_connection', account_id, project_id, connection_id)


@app.command()
def delete_environment(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    environment_id: int = ENVIRONMENT_ID,
):
    """Delete an environment"""
    _dbt_cloud_request(
        ctx, 'delete_environment', account_id, project_id, environment_id
    )


@app.command()
def delete_environment_variables(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Delete environment variables"""
    _dbt_cloud_request(
        ctx,
        'delete_environment_variables',
        account_id,
        project_id,
        json.loads(payload),
    )


@app.command()
def delete_project(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
):
    """Delete a project."""
    _dbt_cloud_request(ctx, 'delete_project', account_id, project_id)


@app.command()
def delete_repository(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    repository_id: int = REPOSITORY_ID,
):
    """Delete a repository."""
    _dbt_cloud_request(ctx, 'delete_repository', account_id, project_id, repository_id)


@app.command()
def delete_user_group(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    group_id: int = GROUP_ID,
):
    """Delete a user group."""
    _dbt_cloud_request(ctx, 'delete_user_group', account_id, group_id)


@app.command()
def get_account(ctx: typer.Context, account_id: int = ACCOUNT_ID):
    """Get an account by its ID."""
    _dbt_cloud_request(ctx, 'get_account', account_id)


@app.command()
def get_account_by_name(
    ctx: typer.Context,
    account_name: str = typer.Option(..., '--account-name', help='Name of account'),
):
    """Get an account by its ID."""
    _dbt_cloud_request(ctx, 'get_account_by_name', account_name)


@app.command()
def get_account_licenses(ctx: typer.Context, account_id: int = ACCOUNT_ID):
    """Get an account by its ID."""
    _dbt_cloud_request(ctx, 'get_account_licenses', account_id)


@app.command()
def get_job(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    job_id: int = JOB_ID,
    order_by: str = ORDER_BY,
):
    """Return job details for a job on an account."""
    _dbt_cloud_request(
        ctx,
        'get_job',
        account_id,
        job_id,
        order_by=order_by,
    )


@app.command()
def get_project(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
):
    """Get a project by its ID"""
    _dbt_cloud_request(ctx, 'get_project', account_id=account_id, project_id=project_id)


@app.command()
def get_project_by_name(
    ctx: typer.Context,
    project_name: str = typer.Option(..., '--project-name', help='Name of project'),
    account_id: str = typer.Option(
        None, '--account-id', '-a', help='Numeric ID of account'
    ),
    account_name: str = typer.Option(None, '--account-name', help='Name of account'),
):
    """Get a project by its name."""
    _dbt_cloud_request(
        ctx,
        'get_project_by_name',
        project_name,
        account_id=account_id,
        account_name=account_name,
    )


@app.command()
def get_run(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    run_id: int = RUN_ID,
    include_related: List[str] = INCLUDE_RELATED,
):
    """Get run by ID for a specific account."""
    _dbt_cloud_request(
        ctx,
        'get_run',
        account_id,
        run_id,
        include_related=include_related,
    )


@app.command()
def get_run_artifact(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    run_id: int = RUN_ID,
    path: str = typer.Option(
        ..., '--path', '-f', help='Name of artifact to retrieve (e.g. manifest.json)'
    ),
    step: int = typer.Option(
        None, '--step', '-s', help='Index of the step in the run to retrieve'
    ),
):
    """Get a specific run artifact by path

    Use this endpoint to fetch artifacts from a completed run.  Once a run has
    completed, you can use this endpoint to download the `manifest.json`,
    `run_results.json`, or `catalog.json` files from dbt Cloud.  These artifacts
    contain information about the models in your dbt project, timing information
    around their execution, and a status message indicating the result of the model
    build

    Note:  By default, this endpoint returns artifacts from the last step in the run.
    To list artifacts from other steps in the run, use the `step` query parameter.
    """
    _dbt_cloud_request(
        ctx,
        'get_run_artifact',
        account_id=account_id,
        run_id=run_id,
        path=path,
        step=step,
    )


@app.command()
def get_run_timing_details(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    run_id: int = RUN_ID,
):
    """Get run timing details for specific run."""
    _dbt_cloud_request(
        ctx,
        'get_run_timing_details',
        account_id,
        project_id,
        run_id,
    )


@app.command()
def get_run_v4(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    run_id: int = RUN_ID,
):
    """Get specific run, from version 4"""
    _dbt_cloud_request(
        ctx,
        'get_run_v4',
        account_id,
        run_id,
    )


@app.command()
def get_user(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    user_id: int = USER_ID,
):
    """Get a user."""
    _dbt_cloud_request(
        ctx,
        'get_user',
        account_id,
        user_id,
    )


@app.command()
def list_accounts(ctx: typer.Context):
    """List of accounts that your API Token is authorized to access."""
    _dbt_cloud_request(ctx, 'list_accounts')


@app.command()
def list_audit_logs(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    logged_at_start: str = typer.Option(
        None, '--logged-at-start', help='Date to begin retrieving audit logs'
    ),
    logged_at_end: str = typer.Option(
        None, '--logged-at-end', help='Date to stop retrieving audit logs'
    ),
    offset: int = OFFSET,
    limit: int = LIMIT,
):
    """Retrieve audit logs for an account."""
    _dbt_cloud_request(
        ctx,
        'list_audit_logs',
        account_id,
        logged_at_start=logged_at_start,
        logged_at_end=logged_at_end,
        offset=offset,
        limit=limit,
    )


@app.command()
def list_connections(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
):
    """List connections for a specific account and project."""
    _dbt_cloud_request(ctx, 'list_connections', account_id, project_id)


@app.command()
def list_credentials(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
):
    """List credentials for a specific account and project."""
    _dbt_cloud_request(ctx, 'list_credentials', account_id, project_id)


@app.command()
def list_environments(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
):
    """List environments for a specific account and project."""
    _dbt_cloud_request(ctx, 'list_environments', account_id, project_id)


@app.command()
def list_feature_flags(ctx: typer.Context, account_id: int = ACCOUNT_ID):
    """List feature flags for a specific account."""
    _dbt_cloud_request(ctx, 'list_feature_flags', account_id)


@app.command()
def list_groups(ctx: typer.Context, account_id: int = ACCOUNT_ID):
    """List groups for a specific account."""
    _dbt_cloud_request(ctx, 'list_groups', account_id)


@app.command()
def list_invited_users(ctx: typer.Context, account_id: int = ACCOUNT_ID):
    """List invited users for a specific account."""
    _dbt_cloud_request(ctx, 'list_invited_users', account_id)


@app.command()
def list_jobs(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = typer.Option(
        None, '--project-id', '-p', help='Numeric ID of the project to retrieve.'
    ),
    order_by: str = ORDER_BY,
):
    """List jobs in an account or specific project"""
    _dbt_cloud_request(
        ctx,
        'list_jobs',
        account_id,
        order_by=order_by,
        project_id=project_id,
    )


@app.command()
def list_projects(ctx: typer.Context, account_id: int = ACCOUNT_ID):
    """List projects for a specified account"""
    _dbt_cloud_request(ctx, 'list_projects', account_id)


@app.command()
def list_repositories(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
):
    """List repositories for a specific account and project."""
    _dbt_cloud_request(ctx, 'list_repositories', account_id, project_id)


@app.command()
def list_run_artifacts(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    run_id: int = RUN_ID,
    step: int = typer.Option(None, '--step', '-s'),
):
    """List run artifacts

    Use this endpoint to fetch a list of artifact fiels generated for a completed run.

    Note:  By default, this endpoint returns artifacts from the last step in the run.
    To list artifacts from other steps in the run, use the `step` query parameter.
    """
    _dbt_cloud_request(ctx, 'list_run_artifacts', account_id, run_id, step=step)


@app.command()
def list_runs(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    include_related: List[str] = INCLUDE_RELATED,
    job_id: int = typer.Option(
        None, '--job-id', '-j', help='Numeric ID of job to retrieve'
    ),
    order_by: str = ORDER_BY,
    offset: int = OFFSET,
    limit: int = LIMIT,
):
    """List runs for a specific account."""
    _dbt_cloud_request(
        ctx,
        'list_runs',
        account_id,
        include_related=include_related,
        job_definition_id=job_id,
        order_by=order_by,
        offset=offset,
        limit=limit,
    )


@app.command()
def list_service_token_permissions(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    service_token_id: int = SERVICE_TOKEN_ID,
):
    """List service token permissions for a specific account."""
    _dbt_cloud_request(
        ctx, 'list_service_token_permissions', account_id, service_token_id
    )


@app.command()
def list_users(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    order_by: str = ORDER_BY,
    offset: int = OFFSET,
    limit: int = LIMIT,
):
    """List users for a specific account."""
    _dbt_cloud_request(
        ctx,
        'list_users',
        account_id,
        order_by=order_by,
        offset=offset,
        limit=limit,
    )


@app.command()
def test_connection(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Test a warehouse connection."""
    _dbt_cloud_request(
        ctx,
        'test_connection',
        account_id,
        project_id,
        json.loads(payload),
    )


@app.command()
def trigger_job(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    job_id: int = JOB_ID,
    payload: str = PAYLOAD,
):
    """Trigger job to run."""
    _dbt_cloud_request(
        ctx,
        'trigger_job',
        account_id,
        job_id,
        json.loads(payload),
    )


@app.command()
def trigger_job_and_poll(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    job_id: int = JOB_ID,
    payload: str = PAYLOAD,
    poll_interval: int = typer.Option(
        None, '--poll-interval', help='Number of seconds to wait in between polling.'
    ),
):
    """Trigger job and poll until completion (success or failure)."""
    _dbt_cloud_request(
        ctx,
        'trigger_job_and_poll',
        account_id,
        job_id,
        json.loads(payload),
        poll_interval=poll_interval,
    )


@app.command()
def update_connection(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    connection_id: int = CONNECTION_ID,
    payload: str = PAYLOAD,
):
    """Update a connection."""
    _dbt_cloud_request(
        ctx,
        'update_connection',
        account_id,
        project_id,
        connection_id,
        json.loads(payload),
    )


@app.command()
def update_credentials(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    credentials_id: int = typer.Option(
        ..., '--credentials-id', help='Numeric ID of the credentials.'
    ),
    payload: str = PAYLOAD,
):
    """Update credentials."""
    _dbt_cloud_request(
        ctx,
        'update_credentials',
        account_id,
        project_id,
        credentials_id,
        json.loads(payload),
    )


@app.command()
def update_environment(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    environment_id: int = ENVIRONMENT_ID,
    payload: str = PAYLOAD,
):
    """Update an environment."""
    _dbt_cloud_request(
        ctx,
        'update_environment',
        account_id,
        project_id,
        environment_id,
        json.loads(payload),
    )


@app.command()
def update_job(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    job_id: int = JOB_ID,
    payload: str = PAYLOAD,
):
    """Update the definition of an existing job."""
    _dbt_cloud_request(
        ctx,
        'update_job',
        account_id,
        job_id,
        json.loads(payload),
    )


@app.command()
def update_project(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    payload: str = PAYLOAD,
):
    """Update a project."""
    _dbt_cloud_request(
        ctx,
        'update_project',
        account_id,
        project_id,
        json.loads(payload),
    )


@app.command()
def update_repository(
    ctx: typer.Context,
    account_id: int = ACCOUNT_ID,
    project_id: int = PROJECT_ID,
    repository_id: int = REPOSITORY_ID,
    payload: str = PAYLOAD,
):
    """Update a repository."""
    _dbt_cloud_request(
        ctx,
        'update_repository',
        account_id,
        project_id,
        repository_id,
        json.loads(payload),
    )


def main():
    app()



# stdlib
from typing import Dict

# third party
from sgqlc.endpoint.http import HTTPEndpoint
from sgqlc.operation import Operation

# first party
from dbtc.client.base import _Client
from dbtc.client.metadata.schema import Query


class _MetadataClient(_Client):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    _header_property = 'service_token'
    _default_domain = 'metadata.cloud.getdbt.com'
    _path = '/graphql'

    @property
    def _endpoint(self) -> HTTPEndpoint:
        return HTTPEndpoint(self.full_url(), self.headers)

    def _make_request(self, obj: str, arguments: Dict = None) -> Dict:
        op = Operation(Query)
        getattr(op, obj)(  # noqa: F841
            **{k: v for k, v in arguments.items() if v is not None}  # type: ignore
        ).__fields__()
        data = self._endpoint(op)
        return data

    def get_exposure(self, job_id: int, name: str, *, run_id: int = None) -> Dict:
        return self._make_request(
            "exposure",
            {"job_id": job_id, "name": name, "run_id": run_id},
        )

    def get_exposures(
        self,
        job_id: int,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "exposures",
            {"job_id": job_id, "run_id": run_id},
        )

    def get_macro(
        self,
        job_id: int,
        unique_id: str,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "macro",
            {"job_id": job_id, "unique_id": unique_id, "run_id": run_id},
        )

    def get_macros(
        self,
        job_id: int,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request("macros", {"job_id": job_id, "run_id": run_id})

    def get_metric(
        self,
        job_id: int,
        unique_id: str,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "metric",
            {"job_id": job_id, "unique_id": unique_id, "run_id": run_id},
        )

    def get_metrics(
        self,
        job_id: int,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request("metrics", {"job_id": job_id, "run_id": run_id})

    def get_model(
        self,
        job_id: int,
        unique_id: str,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "model",
            {"job_id": job_id, "unique_id": unique_id, "run_id": run_id},
        )

    def get_models(
        self,
        job_id: int,
        *,
        database: str = None,
        schema: str = None,
        identifier: str = None,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "models",
            {
                "job_id": job_id,
                "database": database,
                "schema": schema,
                "identifier": identifier,
                "run_id": run_id,
            },
        )

    def get_seed(
        self,
        job_id: int,
        unique_id: str,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "seed",
            {"job_id": job_id, "unique_id": unique_id, "run_id": run_id},
        )

    def get_seeds(
        self,
        job_id: int,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "seeds",
            {"job_id": job_id, "run_id": run_id},
        )

    def get_snapshot(
        self,
        job_id: int,
        unique_id: str,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "snapshot",
            {"job_id": job_id, "unique_id": unique_id, "run_id": run_id},
        )

    def get_snapshots(
        self,
        job_id: int,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "snapshots",
            {
                "job_id": job_id,
                "run_id": run_id,
            },
        )

    def get_source(
        self,
        job_id: int,
        unique_id: str,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "source",
            {"job_id": job_id, "unique_id": unique_id, "run_id": run_id},
        )

    def get_sources(
        self,
        job_id: int,
        *,
        database: str = None,
        schema: str = None,
        identifier: str = None,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "sources",
            {
                "job_id": job_id,
                "database": database,
                "schema": schema,
                "identifier": identifier,
                "run_id": run_id,
            },
        )

    def get_test(
        self,
        job_id: int,
        unique_id: str,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "test", {"job_id": job_id, "unique_id": unique_id, "run_id": run_id}
        )

    def get_tests(
        self,
        job_id: int,
        *,
        run_id: int = None,
    ) -> Dict:
        return self._make_request(
            "tests",
            {
                "job_id": job_id,
                "run_id": run_id,
            },
        )



# third party
import sgqlc.types
import sgqlc.types.datetime

schema = sgqlc.types.Schema()


########################################################################
# Scalars and Enumerations
########################################################################
class AnyScalar(sgqlc.types.Scalar):
    __schema__ = schema


Boolean = sgqlc.types.Boolean

DateTime = sgqlc.types.datetime.DateTime

Float = sgqlc.types.Float


class FreshnessStatus(sgqlc.types.Enum):
    __schema__ = schema
    __choices__ = ('Error', 'Pass', 'Warn')


Int = sgqlc.types.Int


class JSON(sgqlc.types.Scalar):
    __schema__ = schema


class JSONObject(sgqlc.types.Scalar):
    __schema__ = schema


String = sgqlc.types.String


class TimePeriod(sgqlc.types.Enum):
    __schema__ = schema
    __choices__ = ('day', 'hour', 'minute')


########################################################################
# Input Objects
########################################################################

########################################################################
# Output Objects and Interfaces
########################################################################
class CatalogColumn(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('name', 'index', 'type', 'comment', 'description', 'tags')
    name = sgqlc.types.Field(String, graphql_name='name')
    index = sgqlc.types.Field(Int, graphql_name='index')
    type = sgqlc.types.Field(String, graphql_name='type')
    comment = sgqlc.types.Field(String, graphql_name='comment')
    description = sgqlc.types.Field(String, graphql_name='description')
    tags = sgqlc.types.Field(sgqlc.types.list_of(String), graphql_name='tags')


class CatalogStat(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('id', 'label', 'description', 'include', 'value')
    id = sgqlc.types.Field(String, graphql_name='id')
    label = sgqlc.types.Field(String, graphql_name='label')
    description = sgqlc.types.Field(String, graphql_name='description')
    include = sgqlc.types.Field(Boolean, graphql_name='include')
    value = sgqlc.types.Field(AnyScalar, graphql_name='value')


class CloudArtifactInterface(sgqlc.types.Interface):
    __schema__ = schema
    __field_names__ = ('run_id', 'account_id', 'project_id', 'environment_id', 'job_id')
    run_id = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='runId')
    account_id = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='accountId')
    project_id = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='projectId')
    environment_id = sgqlc.types.Field(
        sgqlc.types.non_null(Int), graphql_name='environmentId'
    )
    job_id = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='jobId')


class Criteria(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('warn_after', 'error_after')
    warn_after = sgqlc.types.Field('CriteriaInfo', graphql_name='warnAfter')
    error_after = sgqlc.types.Field('CriteriaInfo', graphql_name='errorAfter')


class CriteriaInfo(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('count', 'period')
    count = sgqlc.types.Field(Int, graphql_name='count')
    period = sgqlc.types.Field(TimePeriod, graphql_name='period')


class MetricFilter(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('field', 'operator', 'value')
    field = sgqlc.types.Field(String, graphql_name='field')
    operator = sgqlc.types.Field(String, graphql_name='operator')
    value = sgqlc.types.Field(String, graphql_name='value')


class NodeInterface(sgqlc.types.Interface):
    __schema__ = schema
    __field_names__ = (
        'resource_type',
        'unique_id',
        'name',
        'description',
        'meta',
        'dbt_version',
        'tags',
    )
    resource_type = sgqlc.types.Field(
        sgqlc.types.non_null(String), graphql_name='resourceType'
    )
    unique_id = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='uniqueId')
    name = sgqlc.types.Field(String, graphql_name='name')
    description = sgqlc.types.Field(String, graphql_name='description')
    meta = sgqlc.types.Field(JSONObject, graphql_name='meta')
    dbt_version = sgqlc.types.Field(String, graphql_name='dbtVersion')
    tags = sgqlc.types.Field(sgqlc.types.list_of(String), graphql_name='tags')


class Query(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = (
        'exposures',
        'exposure',
        'macros',
        'macro',
        'metrics',
        'metric',
        'models',
        'model',
        'seeds',
        'seed',
        'snapshots',
        'snapshot',
        'sources',
        'source',
        'tests',
        'test',
    )
    exposures = sgqlc.types.Field(
        sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('ExposureNode'))),
        graphql_name='exposures',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
            )
        ),
    )
    exposure = sgqlc.types.Field(
        'ExposureNode',
        graphql_name='exposure',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
                (
                    'name',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(String), graphql_name='name', default=None
                    ),
                ),
            )
        ),
    )
    macros = sgqlc.types.Field(
        sgqlc.types.list_of(sgqlc.types.non_null('MacroNode')),
        graphql_name='macros',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
            )
        ),
    )
    macro = sgqlc.types.Field(
        'MacroNode',
        graphql_name='macro',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
                (
                    'unique_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(String),
                        graphql_name='uniqueId',
                        default=None,
                    ),
                ),
            )
        ),
    )
    metrics = sgqlc.types.Field(
        sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('MetricNode'))),
        graphql_name='metrics',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
            )
        ),
    )
    metric = sgqlc.types.Field(
        'MetricNode',
        graphql_name='metric',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
                (
                    'unique_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(String),
                        graphql_name='uniqueId',
                        default=None,
                    ),
                ),
            )
        ),
    )
    models = sgqlc.types.Field(
        sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('ModelNode'))),
        graphql_name='models',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
                (
                    'schema',
                    sgqlc.types.Arg(String, graphql_name='schema', default=None),
                ),
                (
                    'identifier',
                    sgqlc.types.Arg(String, graphql_name='identifier', default=None),
                ),
                (
                    'database',
                    sgqlc.types.Arg(String, graphql_name='database', default=None),
                ),
            )
        ),
    )
    model = sgqlc.types.Field(
        'ModelNode',
        graphql_name='model',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
                (
                    'unique_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(String),
                        graphql_name='uniqueId',
                        default=None,
                    ),
                ),
            )
        ),
    )
    seeds = sgqlc.types.Field(
        sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('SeedNode'))),
        graphql_name='seeds',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
            )
        ),
    )
    seed = sgqlc.types.Field(
        'SeedNode',
        graphql_name='seed',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
                (
                    'unique_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(String),
                        graphql_name='uniqueId',
                        default=None,
                    ),
                ),
            )
        ),
    )
    snapshots = sgqlc.types.Field(
        sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('SnapshotNode'))),
        graphql_name='snapshots',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
            )
        ),
    )
    snapshot = sgqlc.types.Field(
        'SnapshotNode',
        graphql_name='snapshot',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
                (
                    'unique_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(String),
                        graphql_name='uniqueId',
                        default=None,
                    ),
                ),
            )
        ),
    )
    sources = sgqlc.types.Field(
        sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('SourceNode'))),
        graphql_name='sources',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
                (
                    'schema',
                    sgqlc.types.Arg(String, graphql_name='schema', default=None),
                ),
                (
                    'identifier',
                    sgqlc.types.Arg(String, graphql_name='identifier', default=None),
                ),
                (
                    'database',
                    sgqlc.types.Arg(String, graphql_name='database', default=None),
                ),
            )
        ),
    )
    source = sgqlc.types.Field(
        'SourceNode',
        graphql_name='source',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
                (
                    'unique_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(String),
                        graphql_name='uniqueId',
                        default=None,
                    ),
                ),
            )
        ),
    )
    tests = sgqlc.types.Field(
        sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('TestNode'))),
        graphql_name='tests',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
            )
        ),
    )
    test = sgqlc.types.Field(
        'TestNode',
        graphql_name='test',
        args=sgqlc.types.ArgDict(
            (
                (
                    'job_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(Int), graphql_name='jobId', default=None
                    ),
                ),
                ('run_id', sgqlc.types.Arg(Int, graphql_name='runId', default=None)),
                (
                    'unique_id',
                    sgqlc.types.Arg(
                        sgqlc.types.non_null(String),
                        graphql_name='uniqueId',
                        default=None,
                    ),
                ),
            )
        ),
    )


class RunInfoNode(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = (
        'execution_time',
        'thread_id',
        'run_generated_at',
        'run_elapsed_time',
        'error',
        'status',
        'skip',
        'compile_started_at',
        'compile_completed_at',
        'execute_started_at',
        'execute_completed_at',
        'invocation_id',
        'args',
    )  # type: ignore
    execution_time = sgqlc.types.Field(Float, graphql_name='executionTime')
    thread_id = sgqlc.types.Field(String, graphql_name='threadId')
    run_generated_at = sgqlc.types.Field(DateTime, graphql_name='runGeneratedAt')
    run_elapsed_time = sgqlc.types.Field(Float, graphql_name='runElapsedTime')
    error = sgqlc.types.Field(String, graphql_name='error')
    status = sgqlc.types.Field(String, graphql_name='status')
    skip = sgqlc.types.Field(Boolean, graphql_name='skip')
    compile_started_at = sgqlc.types.Field(DateTime, graphql_name='compileStartedAt')
    compile_completed_at = sgqlc.types.Field(
        DateTime, graphql_name='compileCompletedAt'
    )
    execute_started_at = sgqlc.types.Field(DateTime, graphql_name='executeStartedAt')
    execute_completed_at = sgqlc.types.Field(
        DateTime, graphql_name='executeCompletedAt'
    )
    invocation_id = sgqlc.types.Field(String, graphql_name='invocationId')
    args = sgqlc.types.Field(JSON, graphql_name='args')


class ExposureNode(sgqlc.types.Type, CloudArtifactInterface, NodeInterface):
    __schema__ = schema
    __field_names__ = (
        'manifest_generated_at',
        'package_name',
        'owner_email',
        'owner_name',
        'exposure_type',
        'url',
        'maturity',
        'depends_on',
        'parents',
        'parents_sources',
        'parents_models',
    )  # type: ignore
    manifest_generated_at = sgqlc.types.Field(
        DateTime, graphql_name='manifestGeneratedAt'
    )
    package_name = sgqlc.types.Field(String, graphql_name='packageName')
    owner_email = sgqlc.types.Field(String, graphql_name='ownerEmail')
    owner_name = sgqlc.types.Field(String, graphql_name='ownerName')
    exposure_type = sgqlc.types.Field(String, graphql_name='exposureType')
    url = sgqlc.types.Field(String, graphql_name='url')
    maturity = sgqlc.types.Field(String, graphql_name='maturity')
    depends_on = sgqlc.types.Field(
        sgqlc.types.list_of(sgqlc.types.non_null(String)), graphql_name='dependsOn'
    )
    parents = sgqlc.types.Field(
        sgqlc.types.list_of(sgqlc.types.non_null(NodeInterface)), graphql_name='parents'
    )
    parents_sources = sgqlc.types.Field(
        sgqlc.types.list_of('SourceNode'), graphql_name='parentsSources'
    )
    parents_models = sgqlc.types.Field(
        sgqlc.types.list_of('ModelNode'), graphql_name='parentsModels'
    )


class MacroNode(sgqlc.types.Type, CloudArtifactInterface, NodeInterface):
    __schema__ = schema
    __field_names__ = (
        'package_name',
        'path',
        'root_path',
        'original_file_path',
        'depends_on',
        'macro_sql',
    )  # type: ignore
    package_name = sgqlc.types.Field(String, graphql_name='packageName')
    path = sgqlc.types.Field(String, graphql_name='path')
    root_path = sgqlc.types.Field(String, graphql_name='rootPath')
    original_file_path = sgqlc.types.Field(String, graphql_name='originalFilePath')
    depends_on = sgqlc.types.Field(
        sgqlc.types.list_of(String), graphql_name='dependsOn'
    )
    macro_sql = sgqlc.types.Field(String, graphql_name='macroSql')


class MetricNode(sgqlc.types.Type, CloudArtifactInterface, NodeInterface):
    __schema__ = schema
    __field_names__ = (
        'package_name',
        'label',
        'type',
        'sql',
        'timestamp',
        'filters',
        'time_grains',
        'dimensions',
        'depends_on',
        'model',
    )  # type: ignore
    package_name = sgqlc.types.Field(String, graphql_name='packageName')
    label = sgqlc.types.Field(String, graphql_name='label')
    type = sgqlc.types.Field(String, graphql_name='type')
    sql = sgqlc.types.Field(String, graphql_name='sql')
    timestamp = sgqlc.types.Field(String, graphql_name='timestamp')
    filters = sgqlc.types.Field(
        sgqlc.types.list_of(MetricFilter), graphql_name='filters'
    )
    time_grains = sgqlc.types.Field(
        sgqlc.types.list_of(String), graphql_name='timeGrains'
    )
    dimensions = sgqlc.types.Field(
        sgqlc.types.list_of(String), graphql_name='dimensions'
    )
    depends_on = sgqlc.types.Field(
        sgqlc.types.list_of(String), graphql_name='dependsOn'
    )
    model = sgqlc.types.Field('ModelNode', graphql_name='model')


class ModelNode(sgqlc.types.Type, CloudArtifactInterface, NodeInterface):
    __schema__ = schema
    __field_names__ = (
        'database',
        'schema',
        'alias',
        'invocation_id',
        'args',
        'error',
        'status',
        'skip',
        'compile_started_at',
        'compile_completed_at',
        'execute_started_at',
        'execute_completed_at',
        'execution_time',
        'thread_id',
        'run_generated_at',
        'run_elapsed_time',
        'depends_on',
        'package_name',
        'type',
        'owner',
        'comment',
        'children_l1',
        'raw_sql',
        'compiled_sql',
        'columns',
        'stats',
        'run_results',
        'parents_models',
        'parents_sources',
        'tests',
    )  # type: ignore
    database = sgqlc.types.Field(String, graphql_name='database')
    schema = sgqlc.types.Field(String, graphql_name='schema')
    alias = sgqlc.types.Field(String, graphql_name='alias')
    invocation_id = sgqlc.types.Field(String, graphql_name='invocationId')
    args = sgqlc.types.Field(JSON, graphql_name='args')
    error = sgqlc.types.Field(String, graphql_name='error')
    status = sgqlc.types.Field(String, graphql_name='status')
    skip = sgqlc.types.Field(Boolean, graphql_name='skip')
    compile_started_at = sgqlc.types.Field(DateTime, graphql_name='compileStartedAt')
    compile_completed_at = sgqlc.types.Field(
        DateTime, graphql_name='compileCompletedAt'
    )
    execute_started_at = sgqlc.types.Field(DateTime, graphql_name='executeStartedAt')
    execute_completed_at = sgqlc.types.Field(
        DateTime, graphql_name='executeCompletedAt'
    )
    execution_time = sgqlc.types.Field(Float, graphql_name='executionTime')
    thread_id = sgqlc.types.Field(String, graphql_name='threadId')
    run_generated_at = sgqlc.types.Field(DateTime, graphql_name='runGeneratedAt')
    run_elapsed_time = sgqlc.types.Field(Float, graphql_name='runElapsedTime')
    depends_on = sgqlc.types.Field(
        sgqlc.types.list_of(String), graphql_name='dependsOn'
    )
    package_name = sgqlc.types.Field(String, graphql_name='packageName')
    type = sgqlc.types.Field(String, graphql_name='type')
    owner = sgqlc.types.Field(String, graphql_name='owner')
    comment = sgqlc.types.Field(String, graphql_name='comment')
    children_l1 = sgqlc.types.Field(
        sgqlc.types.list_of(sgqlc.types.non_null(String)), graphql_name='childrenL1'
    )
    raw_sql = sgqlc.types.Field(String, graphql_name='rawSql')
    compiled_sql = sgqlc.types.Field(String, graphql_name='compiledSql')
    columns = sgqlc.types.Field(
        sgqlc.types.list_of(CatalogColumn), graphql_name='columns'
    )
    stats = sgqlc.types.Field(sgqlc.types.list_of(CatalogStat), graphql_name='stats')
    run_results = sgqlc.types.Field(
        sgqlc.types.list_of(sgqlc.types.non_null(RunInfoNode)),
        graphql_name='runResults',
    )
    parents_models = sgqlc.types.Field(
        sgqlc.types.list_of(sgqlc.types.non_null('ModelNode')),
        graphql_name='parentsModels',
    )
    parents_sources = sgqlc.types.Field(
        sgqlc.types.list_of(sgqlc.types.non_null('SourceNode')),
        graphql_name='parentsSources',
    )
    tests = sgqlc.types.Field(sgqlc.types.list_of('TestNode'), graphql_name='tests')


class SeedNode(sgqlc.types.Type, CloudArtifactInterface, NodeInterface):
    __schema__ = schema
    __field_names__ = (
        'package_name',
        'database',
        'schema',
        'alias',
        'error',
        'status',
        'skip',
        'compile_started_at',
        'compile_completed_at',
        'execute_started_at',
        'execute_completed_at',
        'execution_time',
        'run_generated_at',
        'run_elapsed_time',
        'columns',
        'stats',
        'thread_id',
        'children_l1',
        'type',
        'owner',
        'comment',
        'raw_sql',
        'compiled_sql',
    )  # type: ignore
    package_name = sgqlc.types.Field(String, graphql_name='packageName')
    database = sgqlc.types.Field(String, graphql_name='database')
    schema = sgqlc.types.Field(String, graphql_name='schema')
    alias = sgqlc.types.Field(String, graphql_name='alias')
    error = sgqlc.types.Field(String, graphql_name='error')
    status = sgqlc.types.Field(String, graphql_name='status')
    skip = sgqlc.types.Field(Boolean, graphql_name='skip')
    compile_started_at = sgqlc.types.Field(DateTime, graphql_name='compileStartedAt')
    compile_completed_at = sgqlc.types.Field(
        DateTime, graphql_name='compileCompletedAt'
    )
    execute_started_at = sgqlc.types.Field(DateTime, graphql_name='executeStartedAt')
    execute_completed_at = sgqlc.types.Field(
        DateTime, graphql_name='executeCompletedAt'
    )
    execution_time = sgqlc.types.Field(Float, graphql_name='executionTime')
    run_generated_at = sgqlc.types.Field(DateTime, graphql_name='runGeneratedAt')
    run_elapsed_time = sgqlc.types.Field(Float, graphql_name='runElapsedTime')
    columns = sgqlc.types.Field(
        sgqlc.types.list_of(CatalogColumn), graphql_name='columns'
    )
    stats = sgqlc.types.Field(sgqlc.types.list_of(CatalogStat), graphql_name='stats')
    thread_id = sgqlc.types.Field(String, graphql_name='thread_id')
    children_l1 = sgqlc.types.Field(
        sgqlc.types.list_of(String), graphql_name='childrenL1'
    )
    type = sgqlc.types.Field(String, graphql_name='type')
    owner = sgqlc.types.Field(String, graphql_name='owner')
    comment = sgqlc.types.Field(String, graphql_name='comment')
    raw_sql = sgqlc.types.Field(String, graphql_name='rawSql')
    compiled_sql = sgqlc.types.Field(String, graphql_name='compiledSql')


class SnapshotNode(sgqlc.types.Type, CloudArtifactInterface, NodeInterface):
    __schema__ = schema
    __field_names__ = (
        'package_name',
        'database',
        'schema',
        'alias',
        'error',
        'status',
        'skip',
        'parents_models',
        'parents_sources',
        'compile_started_at',
        'compile_completed_at',
        'execute_started_at',
        'execute_completed_at',
        'execution_time',
        'thread_id',
        'run_generated_at',
        'run_elapsed_time',
        'columns',
        'stats',
        'children_l1',
        'type',
        'owner',
        'comment',
        'raw_sql',
        'compiled_sql',
    )  # type: ignore
    package_name = sgqlc.types.Field(String, graphql_name='packageName')
    database = sgqlc.types.Field(String, graphql_name='database')
    schema = sgqlc.types.Field(String, graphql_name='schema')
    alias = sgqlc.types.Field(String, graphql_name='alias')
    error = sgqlc.types.Field(String, graphql_name='error')
    status = sgqlc.types.Field(String, graphql_name='status')
    skip = sgqlc.types.Field(Boolean, graphql_name='skip')
    parents_models = sgqlc.types.Field(
        sgqlc.types.list_of(ModelNode), graphql_name='parentsModels'
    )
    parents_sources = sgqlc.types.Field(
        sgqlc.types.list_of('SourceNode'), graphql_name='parentsSources'
    )
    compile_started_at = sgqlc.types.Field(DateTime, graphql_name='compileStartedAt')
    compile_completed_at = sgqlc.types.Field(
        DateTime, graphql_name='compileCompletedAt'
    )
    execute_started_at = sgqlc.types.Field(DateTime, graphql_name='executeStartedAt')
    execute_completed_at = sgqlc.types.Field(
        DateTime, graphql_name='executeCompletedAt'
    )
    execution_time = sgqlc.types.Field(Float, graphql_name='executionTime')
    thread_id = sgqlc.types.Field(String, graphql_name='threadId')
    run_generated_at = sgqlc.types.Field(DateTime, graphql_name='runGeneratedAt')
    run_elapsed_time = sgqlc.types.Field(Float, graphql_name='runElapsedTime')
    columns = sgqlc.types.Field(
        sgqlc.types.list_of(CatalogColumn), graphql_name='columns'
    )
    stats = sgqlc.types.Field(sgqlc.types.list_of(CatalogStat), graphql_name='stats')
    children_l1 = sgqlc.types.Field(
        sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null(String))),
        graphql_name='childrenL1',
    )
    type = sgqlc.types.Field(String, graphql_name='type')
    owner = sgqlc.types.Field(String, graphql_name='owner')
    comment = sgqlc.types.Field(String, graphql_name='comment')
    raw_sql = sgqlc.types.Field(String, graphql_name='rawSql')
    compiled_sql = sgqlc.types.Field(String, graphql_name='compiledSql')


class SourceNode(sgqlc.types.Type, CloudArtifactInterface, NodeInterface):
    __schema__ = schema
    __field_names__ = (
        'database',
        'schema',
        'identifier',
        'source_name',
        'source_description',
        'max_loaded_at',
        'snapshotted_at',
        'max_loaded_at_time_ago_in_s',
        'run_generated_at',
        'run_elapsed_time',
        'state',
        'freshness_checked',
        'criteria',
        'columns',
        'stats',
        'loader',
        'type',
        'owner',
        'comment',
        'children_l1',
        'tests',
    )  # type: ignore
    database = sgqlc.types.Field(String, graphql_name='database')
    schema = sgqlc.types.Field(String, graphql_name='schema')
    identifier = sgqlc.types.Field(String, graphql_name='identifier')
    source_name = sgqlc.types.Field(String, graphql_name='sourceName')
    source_description = sgqlc.types.Field(String, graphql_name='sourceDescription')
    max_loaded_at = sgqlc.types.Field(DateTime, graphql_name='maxLoadedAt')
    snapshotted_at = sgqlc.types.Field(DateTime, graphql_name='snapshottedAt')
    max_loaded_at_time_ago_in_s = sgqlc.types.Field(
        Float, graphql_name='maxLoadedAtTimeAgoInS'
    )
    run_generated_at = sgqlc.types.Field(DateTime, graphql_name='runGeneratedAt')
    run_elapsed_time = sgqlc.types.Field(Float, graphql_name='runElapsedTime')
    state = sgqlc.types.Field(FreshnessStatus, graphql_name='state')
    freshness_checked = sgqlc.types.Field(Boolean, graphql_name='freshnessChecked')
    criteria = sgqlc.types.Field(
        sgqlc.types.non_null(Criteria), graphql_name='criteria'
    )
    columns = sgqlc.types.Field(
        sgqlc.types.list_of(CatalogColumn), graphql_name='columns'
    )
    stats = sgqlc.types.Field(sgqlc.types.list_of(CatalogStat), graphql_name='stats')
    loader = sgqlc.types.Field(String, graphql_name='loader')
    type = sgqlc.types.Field(String, graphql_name='type')
    owner = sgqlc.types.Field(String, graphql_name='owner')
    comment = sgqlc.types.Field(String, graphql_name='comment')
    children_l1 = sgqlc.types.Field(
        sgqlc.types.list_of(sgqlc.types.non_null(String)), graphql_name='childrenL1'
    )
    tests = sgqlc.types.Field(sgqlc.types.list_of('TestNode'), graphql_name='tests')


class TestNode(sgqlc.types.Type, CloudArtifactInterface, NodeInterface):
    __schema__ = schema
    __field_names__ = (
        'state',
        'column_name',
        'status',
        'error',
        'depends_on',
        'fail',
        'warn',
        'skip',
        'raw_sql',
        'compiled_sql',
    )  # type: ignore
    state = sgqlc.types.Field(String, graphql_name='state')
    column_name = sgqlc.types.Field(String, graphql_name='columnName')
    status = sgqlc.types.Field(String, graphql_name='status')
    error = sgqlc.types.Field(String, graphql_name='error')
    depends_on = sgqlc.types.Field(
        sgqlc.types.list_of(String), graphql_name='dependsOn'
    )
    fail = sgqlc.types.Field(Boolean, graphql_name='fail')
    warn = sgqlc.types.Field(Boolean, graphql_name='warn')
    skip = sgqlc.types.Field(Boolean, graphql_name='skip')
    raw_sql = sgqlc.types.Field(String, graphql_name='rawSql')
    compiled_sql = sgqlc.types.Field(String, graphql_name='compiledSql')


########################################################################
# Unions
########################################################################

########################################################################
# Schema Entry Points
########################################################################
schema.query_type = Query
schema.mutation_type = None
schema.subscription_type = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dbtc', 'dbtc.client', 'dbtc.client.cloud', 'dbtc.client.metadata']

package_data = \
{'': ['*'], 'dbtc.client.metadata': ['artifacts/*']}

install_requires = \
['requests>=2.27.1,<3.0.0', 'sgqlc>=15.0,<16.0', 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['dbtc = dbtc.cli:main']}

setup_kwargs = {
    'name': 'dbtc',
    'version': '0.1.1',
    'description': 'An unaffiliated python wrapper for dbt Cloud APIs',
    'long_description': '<p align="center">\n    <em>An unaffiliated python interface for dbt Cloud APIs</em>\n</p>\n<p align="center">\n    <a href="https://codecov.io/gh/dpguthrie/dbtc" target="_blank">\n        <img src="https://img.shields.io/codecov/c/github/dpguthrie/dbtc" alt="Coverage">\n    </a>\n</p>\n\n---\n\n**Documentation**: <a target="_blank" href="https://dbtc.dpguthrie.com">https://dbtc.dpguthrie.com</a>\n\n**Source Code**: <a target="_blank" href="https://github.com/dpguthrie/dbtc">https://github.com/dpguthrie/dbtc</a>\n\n**V2 Docs**: <a target="_blank" href="https://docs.getdbt.com/dbt-cloud/api-v2">https://docs.getdbt.com/dbt-cloud/api-v2</a>\n\n**V3 Docs (Unofficial)**: <a target="_blank" href="https://documenter.getpostman.com/view/14183654/UVsSNiXC">https://documenter.getpostman.com/view/14183654/UVsSNiXC</a>\n\n**V4 Docs**: <a target="_blank" href="https://docs.getdbt.com/dbt-cloud/api-v4">https://docs.getdbt.com/dbt-cloud/api-v4</a>\n\n---\n\n## Overview\n\ndbtc is an unaffiliated python interface to various dbt Cloud API endpoints.\n\nThis library acts as a convenient interface to two different APIs that dbt Cloud offers:\n\n- Cloud API:  This is a REST API that exposes endpoints that allow users to programatically create, read, update, and delete\nresources within their dbt Cloud Account.\n- Metadata API:  This is a GraphQL API that exposes metadata generated from a job run within dbt Cloud.\n\n## Requirements\n\nPython 3.7+\n\n- [Requests](https://requests.readthedocs.io/en/master/) - The elegant and simple HTTP library for Python, built for human beings.\n- [sgqlc]() - Simple GraphQL Client\n- [Typer](https://github.com/ross/requests-futures) - Library for building CLI applications\n\n## Installation\n\n```bash\npip install dbtc\n```\n## Basic Usage\n\n### Python\n\nThe interface to both APIs are located in the `dbtCloudClient` class.\n\nThe example below shows how you use the `cloud` property on an instance of the `dbtCloudClient` class to access methods that allow for programmatic control over dbt Cloud resources.\n\n```python\nfrom dbtc import dbtCloudClient\n\nclient = dbtCloudClient()\n\naccount = client.cloud.get_account_by_name(\'My Account\')\nproject = client.cloud.get_project_by_name(account[\'id\'], \'My Project\')\n\nrun_id = client.cloud.trigger_job_and_poll()\n```\n\n## License\n\nThis project is licensed under the terms of the MIT license.\n',
    'author': 'Doug Guthrie',
    'author_email': 'douglas.p.guthrie@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
